# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


import os
import gtk


def get_resource_directory():
  """Get the resource directory."""
  return os.path.join(os.path.dirname(__file__), 'data')


def get_resource_path(path):
  """Get a resource path relative to the resource directory."""
  return os.path.join(get_resource_directory(), path)


def get_icon_path(name):
  """Get the path to an icon file."""
  return get_resource_path(os.path.join('icons', name))


def load_icon(name):
  """Create an image from an icon name."""
  img = gtk.Image()
  img.set_from_file(get_icon_path(name))
  return img


def load_button(name, tooltip, button_type=gtk.Button):
  """Create an image button from an icon name."""
  button = button_type()
  button.set_image(load_icon(name))
  button.set_tooltip_text(tooltip)
  return button





# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80

import os
import yaml
import gobject
import logbook

log = logbook.Logger('Sessions')


class SessionManager(object):

  def __init__(self, model):
    self.model = model
    self.filename = self.model.home.path('session.yaml')
    self.session = self.load()
    if self.session is None:
      self.session = {'terminals': []}

  def save_session(self):
    self.session['terminals'] = [t.cwd for t in self.model.terminals.items]
    self.model.vim.save_session()
    self.save()
    return True

  def load(self):
    if not os.path.exists(self.filename):
      return
    with open(self.filename) as f:
      data = f.read()
    if not data:
      log.error('Empty file')
      return
    session = yaml.load(data)
    if not isinstance(session, dict):
      log.error('Bad file')
      return
    return session

  def start(self):
    if not self.model.config.get('session', True):
      self.model.terminals.execute()
      return
    terminals = self.session.get('terminals', [])
    if terminals:
      for cwd in terminals:
        self.model.terminals.execute(cwd=cwd)
    else:
      self.model.terminals.execute()
    gobject.timeout_add(5000, self.save_session)

  def save(self):
    data = yaml.dump(self.session, default_flow_style=False)
    with open(self.filename, 'w') as f:
      f.write(data)





# -*- coding: utf-8 -*- 
# (c) 2005-2011 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


import gtk
from a8 import resources


class Action(object):
  """Doing something in the IDE."""
  def __init__(self, key, label, icon):
    self.key = key
    self.label = label
    self.icon = icon

  def create_menuitem(self):
    """Create a menu item from an action."""
    item = gtk.ImageMenuItem()
    item.set_label(self.label)
    item.set_image(resources.load_icon(self.icon))
    item.set_always_show_image(True)
    item.set_data('action_key', self.key)
    return item


def create_action_menu(actions, callback):
  """Create a menu from actions."""
  menu = gtk.Menu()
  for action in actions:
    if action is not None:
      menuitem = action.create_menuitem()
      menuitem.connect('activate', callback)
    else:
      menuitem = gtk.SeparatorMenuItem()
    menu.append(menuitem)
  menu.show_all()
  return menu


class ActionRegistry(dict):
  """Registry of actions."""

  def add(self, action):
    self[action.key] = action

  def activate(self, key, **kw):
    self[key](**kw)




# -*- coding: utf-8 -*- 
# (c) 2005-2011 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80

"""Example a8 extension."""


from datetime import datetime

from a8 import lists


class SaveLogItem(lists.ListItem):

  MARKUP_TEMPLATE = '<b>{0}</b>\n{1}'

  def __init__(self, name):
    self.name = name
    self.time = datetime.now()

  @property
  def markup_args(self):
    return (self.name, self.time)


class SaveLog(lists.ListView):

  ICON = 'lorry.png'

  def save(self, filename):
    self.items.append(SaveLogItem(filename))


def annotate_file(item):
  """Annotate files with the first letter of their name."""
  item.annotation = item.basename[0] + ' '


def setup(app):
  # Simple signal handler to annotate files
  app.connect('file-item-added', annotate_file)
  # Plugin list item
  save_log = SaveLog(app)
  app.connect('file-saved', save_log.save)
  save_log.items.append(SaveLogItem('Started'))
  app.ui.plugins.add_tab(save_log)




# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


import gtk

from pygtkhelpers import delegates, utils

from a8 import shortcuts, resources, version


gtk.window_set_default_icon(resources.load_icon('a8.png').get_pixbuf())


class ApplicationWindow(delegates.WindowView):
  """Main application window."""

  def create_ui(self):
    """Create the user interface."""
    self.stack = gtk.VBox()
    self.widget.add(self.stack)
    self.hpaned = gtk.HPaned()
    self.stack.pack_end(self.hpaned)
    self.vpaned = gtk.VPaned()
    self.hpaned.pack2(self.vpaned)
    self.hpaned.set_position(200)
    self.plugins = PluginTabs()
    self.hpaned.pack1(self.plugins.widget)
    self.plugins.add_main(self.model.buffers)
    self.plugins.add_tab(self.model.files)
    self.plugins.add_tab(self.model.bookmarks)
    self.plugins.add_tab(self.model.terminals)
    self.vpaned.pack1(self.model.vim.widget, resize=True, shrink=False)
    self.vpaned.pack2(self.model.terminals.book, resize=False, shrink=False)
    self.widget.set_title('Abominade loves you.')
    self.accel_group = self.model.shortcuts.create_group()
    self.widget.add_accel_group(self.accel_group)
    self.set_title('')
    self.widget.set_default_size(800, 600)
    self.widget.show_all()

  def on_widget__delete_event(self, window, event):
    self.model.stop()
    return True

  def start(self):
    self.show_and_run()

  def set_title(self, filename):
    self.widget.set_title(u'a8♥u {0}'.format(filename))

  def focus_files(self):
    self.plugins.focus_delegate(self.model.files)

  def focus_bookmarks(self):
    self.plugins.focus_delegate(self.model.bookmarks)

  def focus_terminals(self):
    self.plugins.focus_delegate(self.model.terminals)

  def focus_buffers(self):
    self.models.buffers.items.grab_focus()


class PluginTabs(delegates.SlaveView):
  """Tabs containing the main plugins."""

  TAB_POS = gtk.POS_TOP

  def create_ui(self):
    self.stack = gtk.VPaned()
    self.widget.add(self.stack)
    self.book = gtk.Notebook()
    self.book.set_tab_pos(self.TAB_POS)
    self.stack.pack2(self.book, resize=True, shrink=False)

  def add_main(self, delegate):
    self.stack.pack1(delegate.widget, resize=True, shrink=False)

  def add_tab(self, delegate):
    self.book.append_page(delegate.widget, delegate.create_tab_widget())

  def focus_delegate(self, delegate):
    self.book.set_current_page(self.book.page_num(delegate.widget))
    delegate.items.grab_focus()




# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


"""Buffer list."""


import cgi
import gtk, gtk.gdk
from pygtkhelpers import delegates
from pygtkhelpers.ui import objectlist

from a8 import resources, actions


class ListItem(object):

  MARKUP_TEMPLATE = ''

  @property
  def markup_args(self):
    return ()

  @property
  def markup(self):
    return self.MARKUP_TEMPLATE.format(
      *(cgi.escape(str(i)) for i in self.markup_args))


class ListView(delegates.SlaveView):

  LABEL   = 'Unnamed'
  ICON    = 'application_xp_terminal.png'
  COLUMNS = [objectlist.Column('markup', use_markup=True)]

  TOOL_ACTIONS = []

  def create_ui(self):
    """Create the user interface."""
    self.stack = gtk.VBox()
    self.widget.add(self.stack)
    self.scroll = gtk.ScrolledWindow()
    self.scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
    self.stack.pack_end(self.scroll)
    self.items = objectlist.ObjectList(self.COLUMNS)
    self.items.set_headers_visible(False)
    self.scroll.add(self.items)

  def create_tab_widget(self):
    """Create the tab widget."""
    vb = gtk.VBox()
    vb.set_spacing(3)
    self.tab_icon = resources.load_icon(self.ICON)
    self.tab_icon.set_tooltip_text(self.LABEL)
    self.tab_holder = gtk.EventBox()
    self.tab_holder.add(self.tab_icon)
    self.tab_icon.set_can_focus(False)
    self.tab_holder.set_can_focus(False)
    self.tab_holder.add_events(gtk.gdk.BUTTON_PRESS_MASK)
    self.tab_holder.connect('button-press-event', self.on_tab_icon_button)
    vb.pack_start(self.tab_holder, expand=False)
    vb.show_all()
    return vb

  def create_tool_menu(self):
    return actions.create_action_menu(self.TOOL_ACTIONS,
                                      self.on_tool_activate)

  def on_tool_activate(self, item):
    action_key = item.get_data('action_key')
    callback = getattr(self, 'on_%s_activate' % action_key, None)
    if callback is None:
      raise NotImplementedError(action_key)
    else:
      callback()

  def on_tab_icon_button(self, icon, event):
    if event.button == 3 and self.TOOL_ACTIONS:
      menu = self.create_tool_menu()
      menu.show_all()
      menu.popup(None, None, None, event.button, event.time)



# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


"""YAML-based configuration."""


import os
import argparse
import yaml


class ConfigError(ValueError):
  """An error in the configuration."""


class Config(object):
  """Configuration object"""
  def __init__(self):
    self.opts = {}

  def get(self, k, default=None):
    return self.opts.get(k, default)

  def load_from_file(self, path):
    """Load options from a file."""
    if os.path.exists(path):
      with open(path) as f:
        opts = yaml.load(f)
        self.load_from(opts)

  def load_from(self, opts):
    """Load options from a dict-like or list of pairs."""
    if not opts:
      return
    try:
      self.opts.update(opts)
    except ValueError:
      raise ConfigError('Options are not a k/v map.')


class InstanceDirectory(object):
  """Where user data is stored."""

  def __init__(self, user_path=os.path.expanduser('~/.a8')):
    self.user_path = user_path
    self.create()
    self.config_path = self.path('config.yaml')
    self.shortcuts_path = self.path('shortcuts.yaml')

  def create(self):
    if not os.path.exists(self.user_path):
      os.mkdir(self.user_path)

  def path(self, filename):
    return os.path.join(self.user_path, filename)

  def load_config(self):
    config = Config()
    config.load_from_file(self.config_path)
    return config

  def load_shortcuts(self):
    config = Config()
    config.load_from_file(self.shortcuts_path)
    return config



# -*- coding: utf-8 -*- 
# (c) 2005-2011 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


"""Main Abominade entry point script."""


from a8 import app


def main():
  """Run Abominade."""
  a8 = app.Abominade()
  a8.start()


if __name__ == '__main__':
  main()



# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


import os
import gtk
import logbook

from a8 import actions, resources, config, terminals


log = logbook.Logger('shortcuts')


commands = [
  actions.Action('shell', 'Shell', 'application_xp_terminal.png'),
  actions.Action('browse_home', 'Browse home directory', 'house_go.png'),
  actions.Action('config', 'Edit configuration', 'cog.png'),
  None,
  actions.Action('close_all_buffers', 'Close all buffers', 'cross.png'),
]


shortcuts = {
  '<Alt>t': 'shell',
  '<Alt>e': 'focus_vim',
  '<Alt>r': 'focus_terminal',
  '<Alt>b': 'focus_buffers',
  '<Alt>m': 'focus_files',
  '<Alt>i': 'focus_terminals',
  '<Alt>k': 'focus_bookmarks',
  '<Alt>Up': 'prev_buffer',
  '<Alt>Down': 'next_buffer',
  '<Alt>Left': 'prev_terminal',
  '<Alt>Right': 'next_terminal',
  '<Alt>g': 'refresh_files',
  '<Alt>c': 'close_all_buffers',
  '<Alt>h': 'browse_home',
}


class CustomShortcut(object):

  def __init__(self):
    self.cmd = None
    self.cwd = None
    self.env = None


class ShortcutManager(object):

  def __init__(self, model):
    self.model = model
    self.config = self.model.home.load_shortcuts()

  def create_group(self):
    accel_group = gtk.AccelGroup()
    for shortcut, action in shortcuts.items():
      shortcut = self.config.opts.get(action, shortcut)
      keyval, modifier = gtk.accelerator_parse(shortcut)
      accel_group.connect_group(keyval, modifier, gtk.ACCEL_VISIBLE,
                                self.accel_callback(action))
    for shortcut, custom in self.get_customs():
      keyval, modifier = gtk.accelerator_parse(shortcut)
      accel_group.connect_group(keyval, modifier, gtk.ACCEL_VISIBLE,
                                self.custom_callback(custom))
    return accel_group

  def get_customs(self):
    custom_data = self.config.get('custom', None)
    if not custom_data or not isinstance(custom_data, list):
      return
    for custom_datum in custom_data:
      if not isinstance(custom_datum, dict):
        continue
      shortcut = custom_datum.get('key')
      if not shortcut:
        continue
      custom = CustomShortcut()
      custom.cmd = custom_datum.get('cmd')
      custom.cwd = custom_datum.get('cwd')
      custom.env = custom_datum.get('env')
      yield shortcut, custom

  def create_tools(self):
    bar = gtk.HBox()
    for action in commands:
      if action is None:
        bar.pack_start(gtk.Alignment())
      else:
        button = resources.load_button(action.icon, action.label)
        bar.pack_start(button, expand=False)
        button.connect('clicked', self.on_button, action.key)
    bar.show_all()
    return bar

  def on_button(self, button, action_key):
    self.activate(action_key)

  def accel_callback(self, action_key):
    def on_accel(group, acceleratable, keyval, modifier,
                 self=self, action_key=action_key):
      try:
        self.activate(action_key)
      except NotImplementedError, e:
        log.error('NotImplemented: {0}', e)
      except Exception, e:
        raise
        log.error('Unhandled: {0}', e)
      return True
    return on_accel

  def custom_callback(self, custom):
    def on_accel(group, acceleratable, keyval, modifier,
                 self=self, custom=custom):
      self.model.terminals.execute(
        [terminals.get_default_shell(), '-c', custom.cmd],
        custom.cwd, custom.env)
      return True
    return on_accel

  def activate(self, action_key):
    callback = getattr(self, 'on_%s_activate' % action_key, None)
    if callback is None:
      raise NotImplementedError(action_key)
    else:
      callback()

  def on_shell_activate(self):
    self.model.terminals.execute()

  def on_focus_vim_activate(self):
    self.model.vim.grab_focus()

  def on_focus_terminal_activate(self):
    self.model.terminals.grab_focus()

  def on_prev_buffer_activate(self):
    self.model.vim.prev()

  def on_next_buffer_activate(self):
    self.model.vim.next()

  def on_prev_terminal_activate(self):
    self.model.terminals.prev()

  def on_next_terminal_activate(self):
    self.model.terminals.next()

  def on_focus_buffers_activate(self):
    self.model.ui.focus_buffers()

  def on_focus_bookmarks_activate(self):
    self.model.ui.focus_bookmarks()

  def on_focus_files_activate(self):
    self.model.ui.focus_files()

  def on_focus_terminals_activate(self):
    self.model.ui.focus_terminals()

  def on_refresh_files_activate(self):
    self.model.files.on_refresh_activate()

  def on_close_all_buffers_activate(self):
    self.model.vim.close_all()

  def on_browse_home_activate(self):
    self.model.files.browse(os.path.expanduser('~'))

  def on_config_activate(self):
    self.model.vim.open_file(self.model.home.config_path)



# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


"""A library to embed vim in a gtk socket."""


import os, subprocess

import gtk, gobject
import logbook
from pygtkhelpers import delegates

from a8 import bus, resources


log = logbook.Logger('Vim')


class VimManager(delegates.SlaveView):
  """Embed and communicate with Vim."""

  null_callback = {
    'reply_handler': lambda *args: None,
    'error_handler': lambda *args: None,
  }

  def create_ui(self):
    self.holder = gtk.EventBox()
    self.holder.add_events(gtk.gdk.KEY_PRESS_MASK)
    self.widget.add(self.holder)
    self.widget.set_size_request(200, 200)
    self.socket = gtk.Socket()
    self.holder.add(self.socket)
    self.widget.show_all()
    self.vim_command = 'gvim'
    self.vim = None
    self.vim_proc = None
    self.vim_pid = None

  def on_holder__button_press_event(self, eventbox, event):
    self.grab_focus()

  def connect_vim(self):
    log.debug('Connect')
    self.vim = bus.connect('vim')
    self.connect_vim_signals()

  def save_session(self):
    self.vim.command('mks! {0}'.format(self.get_vim_session()))

  def stop(self):
    self.grab_focus()
    self.vim.quit(**self.null_callback)

  def get_vim_env(self):
    env = os.environ.copy()
    env['PYTHONPATH'] = os.path.dirname(os.path.dirname(__file__))
    return env

  def get_vim_args(self):
    args = [
      self.vim_command,
      '--cmd', 'let A8_EMBEDDED=1',
      '--cmd', 'let A8_UID="{uid}"'.format(uid=bus.A8_UID),
      '--cmd', 'so {script}'.format(script=self.get_vim_script()),
      '--socketid', str(self.xid),
    ]
    if self.model.config.get('session', True):
      args.extend(['-S', self.get_vim_session()])
    return args

  def get_vim_session(self):
    path = self.model.home.path('a8_vim_session.vim')
    if not os.path.exists(path):
      f = open(path, 'w')
      f.close()
    return path

  def get_vim_script(self):
    return resources.get_resource_path('a8.vim')

  def execute_vim(self):
    self.vim_proc = subprocess.Popen(
      self.get_vim_args(),
      env=self.get_vim_env(),
      close_fds=True
    )

  def start(self):
    log.debug('Start')
    self.xid = self.socket.get_id()
    if self.vim_pid is not None:
      log.warn('Already started at pid={0}'.format(self.vim_pid))
    try:
      self.execute_vim()
      self.widget.show_all()
      self.connect_vim()
    except OSError, e:
      pass #XXX Warn

  def grab_focus(self):
    gobject.idle_add(self.holder.child_focus, gtk.DIR_TAB_FORWARD)

  def open_file(self, filename):
    self.vim.open_file(filename, **self.null_callback)

  def close(self, filename):
    self.grab_focus()
    self.vim.close_buffer(filename, **self.null_callback)

  def close_all(self):
    self.grab_focus()
    self.vim.close_all_buffers(**self.null_callback)

  def close_under(self, path):
    self.grab_focus()
    self.vim.close_buffers_under(path)

  def goto_line(self, line):
    self.vim.goto_line(line, **self.null_callback)

  def next(self):
    self.vim.next(**self.null_callback)

  def prev(self):
    self.vim.prev(**self.null_callback)

  def connect_vim_signals(self):
    for k in dir(self):
      if k.startswith('onvim_'):
        self.vim.connect_to_signal(k[6:], getattr(self, k))

  def onvim_VimEnter(self):
    log.debug('Signal: Enter')
    for filename in self.model.args.files:
      self.open_file(filename)
    self.grab_focus()

  def onvim_BufEnter(self, bufid, filename):
    self.onvim_BufAdd(bufid, filename)

  def onvim_BufAdd(self, bufid, filename):
    log.debug('Signal: Buffer {0} {1}'.format(bufid, filename))
    path = unicode(filename)
    if not path or os.path.isdir(path):
      path = ''
    else:
      if not os.path.isabs(path):
        cwd = self.vim.get_cwd()
        path = os.path.join(cwd, path)
      path = os.path.realpath(path)
      bufid = int(bufid)
      self.model.buffers.append(path, bufid)
    self.model.emit('file-opened', filename=filename)

  def onvim_BufNew(self, bufid):
    log.debug('New {0}', bufid)

  def onvim_BufDelete(self, bufid, filename):
    path = unicode(filename)
    if not path:
      return
    path = os.path.realpath(path)
    self.model.buffers.remove(path)
    self.model.emit('file-closed', filename=filename)

  def onvim_VimLeave(self):
    log.debug('Signal: Leave')
    gtk.main_quit()

  def onvim_BufWritePost(self, bufid, filename):
    self.model.emit('file-saved', filename=filename)





# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


"""Abominade Monolith."""

import collections
import argparse

from a8 import (terminals, files, buffers, vimembed, window, config, bookmarks,
                shortcuts, extensions, sessions)

class Abominade(object):
  """Abominade Monolith"""

  def __init__(self):
    self.signals = collections.defaultdict(list)
    self.home = config.InstanceDirectory()
    self.config = self.home.load_config()
    self.parse_args()
    self.shortcuts = shortcuts.ShortcutManager(self)
    self.files = files.FileManager(self)
    self.buffers = buffers.BufferManager(self)
    self.terminals = terminals.TerminalManager(self)
    self.bookmarks = bookmarks.BookmarkManager(self)
    self.vim = vimembed.VimManager(self)
    self.ui = window.ApplicationWindow(self)
    self.ui.widget.maximize()
    self.sessions = sessions.SessionManager(self)
    self.sessions.start()
    extensions.load_extensions(self)

  def parse_args(self):
    parser = argparse.ArgumentParser()
    parser.add_argument('files', nargs='*', help='Files to open.')
    parser.add_argument('--no-session', action='store_true')
    parser.add_argument('--show-toolbar', action='store_true')
    self.args = parser.parse_args()
    if self.args.no_session:
      self.config.opts['session'] = False
    if self.args.show_toolbar:
      self.config.opts['toolbar'] = True

  def start(self):
    """Start a8"""
    self.vim.start()
    self.files.browse()
    self.ui.start()

  def stop(self):
    """Stop a8"""
    self.sessions.save_session()
    self.vim.stop()

  def emit(self, signal, **kw):
    for callback in self.signals[signal]:
      callback(**kw)

  def connect(self, signal, callback):
    self.signals[signal].append(callback)



# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


"""Buffer list."""


import os

import gtk, gtk.gdk
from pygtkhelpers.ui import objectlist

from a8 import resources, lists, contexts


class Buffer(lists.ListItem):
  """Loaded buffer."""

  MARKUP_TEMPLATE = '<b>{0}</b>\n<span size="x-small">{1}</span>'

  def __init__(self, model, filename, bufid):
    self.model = model
    self.filename = filename
    self.bufid = bufid
    self.dirname = os.path.dirname(filename)
    self.basename = os.path.basename(filename)
    bookmark = self.model.bookmarks.shortest_path(filename)
    if bookmark:
      supname = self.dirname.replace(bookmark.target, '').lstrip('/')
      self.dispname = '{0}:{1}'.format(bookmark.basename, supname)
    else:
      self.dispname = self.dirname

  @property
  def markup_args(self):
    """Display in the buffer list."""
    return (self.basename, self.dispname)


class BufferManager(lists.ListView):
  """Buffer list."""

  LABEL = 'Buffers'
  ICON  = 'page_white_stack.png'

  COLUMNS = [objectlist.Column('markup', use_markup=True),
             objectlist.Column('bufid', visible=False),
             objectlist.Column('basename', visible=False, searchable=True)]

  def create_ui(self):
    lists.ListView.create_ui(self)
    if self.model.config.get('toolbar', False):
      self.stack.pack_start(self.model.shortcuts.create_tools(), expand=False)
    self.filenames = {}
    self.items.sort_by('bufid')

  def append(self, filename, bufid):
    if filename not in self.filenames:
      self.filenames[filename] = buf = Buffer(self.model, filename, bufid)
      self.items.append(buf)
    if (self.items.get_selection() is None or
        not self.items.selected_item or
        self.items.selected_item.filename != filename):
      b = self.items.selected_item = self.filenames[filename]
      self.model.ui.set_title('{0}/{1}'.format(b.dispname, b.basename))

  def remove(self, filename):
    if filename in self.filenames:
      buf = self.filenames.pop(filename)
      self.items.remove(buf)

  def on_items__item_activated(self, items, item):
    self.model.vim.open_file(item.filename)
    self.model.vim.grab_focus()

  def on_items__item_right_clicked(self, items, item, event):
    context = contexts.LocalContext(self.model, None, item.filename)
    context.create_menu()   # force expanding path, etc.
    # only files should be here, and we want a menu even if file was deleted
    menu = context.create_file_menu()
    if menu is not None:
      menu.popup(None, None, None, event.button, event.time)




# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


import os

import gtk, dbus

# Configure our mainloop
from dbus.mainloop.glib import DBusGMainLoop
mainloop = DBusGMainLoop(set_as_default=True)


A8_UID = 'a8_{uid}'.format(uid=os.getpid())
NS_TEMPLATE = 'uk.co.pida.{name}'
BUS_TEMPLATE = 'uk.co.pida.{name}.{uid}'
PATH_TEMPLATE = '/{name}'


def get_ns(name):
  return NS_TEMPLATE.format(name=name)


def get_busname(name, uid=A8_UID):
  return BUS_TEMPLATE.format(name=name, uid=uid)


def get_path(name):
  return PATH_TEMPLATE.format(name=name)


def create_session(busname):
  def on_watch(busname):
    if busname:
      gtk.main_quit()
  session = dbus.SessionBus()
  watch = session.watch_name_owner(busname, on_watch)
  gtk.main()
  return session


def connect(name):
  busname = get_busname(name)
  session = create_session(busname)
  busobj = session.get_object(busname, get_path(name))
  return dbus.Interface(busobj, get_ns(name))




# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


import os, re
import gtk
from pygtkhelpers.ui import objectlist
from pygtkhelpers import gthreads
gthreads.initial_setup()

from a8 import lists, resources, contexts, actions


HIDDEN_MASKS = [
    '^\.+',
    '.+\.pyc$'
]
HIDDEN_PATTERN = re.compile('|'.join(HIDDEN_MASKS))


class File(lists.ListItem):

  folder_pixbuf = resources.load_icon('folder.png').get_pixbuf()
  file_pixbuf = resources.load_icon('page_white.png').get_pixbuf()

  MARKUP_TEMPLATE = '<b><tt>{0}</tt></b>{1}'

  def __init__(self, filename):
    self.filename = unicode(filename)
    self.basename = os.path.basename(self.filename)
    self.lowname = self.basename.lower()
    self.isdir = os.path.isdir(self.filename)
    self.isdir_key = (self.isdir and 'a' or 'b', self.lowname)
    self.hidden = HIDDEN_PATTERN.match(self.basename)
    self.annotation = ''

  @property
  def markup_args(self):
    return (self.annotation, self.basename)

  @property
  def icon(self):
    if self.isdir:
      return self.folder_pixbuf
    else:
      return self.file_pixbuf


class FileManager(lists.ListView):
  """Files list."""

  LABEL = 'Files'

  ICON  = 'folder.png'
  COLUMNS = [
    objectlist.Column('icon', type=gtk.gdk.Pixbuf),
    objectlist.Column('markup', use_markup=True, searchable=True),
  ]

  TOOL_ACTIONS = [
    actions.Action('refresh', 'Refresh', 'arrow_refresh_small.png')
  ]

  def create_ui(self):
    lists.ListView.create_ui(self)
    self.items.sort_by('isdir_key')

  def browse(self, path=os.getcwd()):
    self.items.clear()
    path = os.path.normpath(path)
    self.path = path
    task = gthreads.GeneratorTask(self.browse_work, self.browse_item)
    task.start(path)
    self.model.ui.focus_files()

  def browse_work(self, path):
    parent = File(os.path.dirname(path))
    parent.basename = '..'
    parent.isdir_key = '1', '_'
    yield parent
    for filename in os.listdir(path):
      yield File(os.path.join(path, filename))

  def browse_item(self, item):
    if not item.hidden:
      self.items.append(item)
      self.model.emit('file-item-added', item=item)

  def on_items__item_activated(self, items, item):
    if item.isdir:
      self.browse(item.filename)
    else:
      self.model.vim.open_file(item.filename)
      self.model.vim.grab_focus()

  def on_items__item_right_clicked(self, items, item, event):
    context = contexts.LocalContext(self.model, None, item.filename)
    menu = context.create_menu()
    menu.popup(None, None, None, event.button, event.time)

  def on_refresh_activate(self):
    self.browse(self.path)




# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


"""Application contexts."""


import os
import gtk
import psutil
import logbook
import webbrowser
from a8 import actions


log = logbook.Logger('Contexts')
log.debug('Log start')


class BaseContext(object):
  """Base context superclass."""

  #: A regular expression to match the context
  expr = None
  name = 'Unnamed Context'

  def __init__(self, model, view, data):
    self.model = model
    self.data = data
    self.view = view

  def create_action_menu(self, acts):
    """Create a menu for the context."""
    return actions.create_action_menu(acts, self.on_menuitem_activate)

  def on_menuitem_activate(self, item):
    action_key = item.get_data('action_key')
    callback = getattr(self, 'on_%s_activate' % action_key, None)
    if callback is None:
      raise NotImplementedError(action_key)
    else:
      callback()


class LocalContext(BaseContext):
  """Context for files and directories."""

  expr = '[^\n :\$]+'
  name = 'Local filesystem context'

  dir_actions = [
    actions.Action(
      'browse_dir',
      'Browse Directory',
      'folder.png'
    ),
    actions.Action(
      'shell_dir',
      'Terminal in Directory',
      'application_xp_terminal.png'
    ),
    actions.Action(
      'bookmark',
      'Add bookmark',
      'star.png'
    ),
    None,
    actions.Action(
      'close_under',
      'Close any children',
      'cross.png',
    ),
  ]

  file_actions = [
    actions.Action(
      'open_file',
      'Open file',
      'page_white.png'
    ),
    None,
    actions.Action(
      'browse_file',
      'Browse parent',
      'folder.png',
    ),
    actions.Action(
      'shell_file',
      'Shell in parent',
      'application_xp_terminal.png',
    ),
    actions.Action(
      'bookmark',
      'Add bookmark',
      'star.png'
    )
  ]

  open_file_actions = [
    None, # Separator
    actions.Action(
      'close_document',
      'Close document',
      'cross.png',
    )
  ]

  def create_menu(self):
    """Create a menu for the context."""
    self.data = os.path.expanduser(self.data)
    if not os.path.isabs(self.data) and self.view is not None:
      log.debug('relative to "{0}"', self.view.cwd)
      self.data = os.path.join(self.view.cwd, self.data)
    if os.path.isdir(self.data):
      log.debug('directory')
      return self.create_dir_menu()
    elif os.path.exists(self.data):
      log.debug('file')
      return self.create_file_menu()

  def create_dir_menu(self):
    return self.create_action_menu(self.dir_actions)

  def create_file_menu(self):
    actions = list(self.file_actions)
    if self.data in self.model.buffers.filenames:
      actions.extend(self.open_file_actions)
    return self.create_action_menu(actions)

  def on_browse_dir_activate(self):
    self.model.files.browse(self.data)

  def on_browse_file_activate(self):
    self.model.files.browse(os.path.dirname(self.data))

  def on_shell_dir_activate(self):
    self.model.terminals.execute(cwd=self.data)

  def on_shell_file_activate(self):
    self.model.terminals.execute(cwd=os.path.dirname(self.data))

  def on_bookmark_activate(self):
    self.model.bookmarks.add(self.data)

  def on_close_document_activate(self):
    self.model.vim.close(self.data)

  def on_open_file_activate(self):
    self.model.vim.open_file(self.data)

  def on_close_under_activate(self):
    self.model.vim.close_under(self.data)


class UriContext(BaseContext):
  """Context for URIs."""

  name = 'URI Context'
  expr = r'https{0,1}://\S+'

  uri_actions = [
    actions.Action('browse_uri', 'Open', 'world_go.png'),
  ]

  def on_browse_uri_activate(self):
    webbrowser.open_new_tab(self.data)

  def create_menu(self):
    return self.create_action_menu(self.uri_actions)


class IntegerContext(BaseContext):
  """Context for integers."""

  expr = r'[0-9]+'
  name = 'Integer context'

  int_actions = [
    actions.Action('term', 'SIGTERM', 'asterisk_yellow.png'),
    actions.Action('kill', 'SIGKILL', 'asterisk_orange.png'),
  ]

  def create_menu(self):
    try:
      pid = int(self.data)
      self.proc = psutil.Process(pid)
      menu = self.create_action_menu(self.int_actions)
    except (ValueError, psutil.NoSuchProcess):
      menu = None
    return menu

  def on_kill_activate(self):
    self.proc.kill()

  def on_term_activate(self):
    self.proc.terminate()


class ContextManager(object):
  context_order = [UriContext, IntegerContext, LocalContext]




# -*- coding: utf-8 -*- 
# (c) 2005-2011 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


from setuptools import setup
from a8 import version


setup(
  name = 'a8',
  author = 'Ali Afshar',
  author_email = 'aafshar@gmail.com',
  description = 'The Abominade IDE',
  license = 'GPL',
  url = 'http://abominade.org',
  version = version.VERSION,
  install_requires = [
    'psutil',
    'logbook',
    'pygtkhelpers>=0.4.3',
    'argparse',
    'pyyaml',
  ],
  packages = [
    'a8',
  ],
  package_data = {
    'a8': [
      'data/icons/*.png',
      'data/a8.vim',
    ],
  },
  entry_points = {
    'console_scripts': [
      'a8 = a8.__main__:main'
    ],
  },
  zip_safe = False,
)



# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80

"""Terminal emulator."""

import os, re, pwd

import logbook
import gtk, gobject, gtk.gdk, pango, vte
import psutil, psutil.error
from pygtkhelpers import delegates
from pygtkhelpers.ui import dialogs

from a8 import resources, lists, contexts


log = logbook.Logger('Terminal')


def get_default_shell():
  """Returns the default shell for the user"""
  # Environ, or fallback to login shell
  return os.environ.get('SHELL', pwd.getpwuid(os.getuid())[-1])


class Unset(object):
  """Indicates an unset, i.e. default, option."""


class TerminalConfiguration(object):
  """Configures a terminal."""

  default_opts = {
    'color_foreground': Unset,
    'color_background': Unset,
    'backspace_binding': Unset,
    'cursor_blink_mode': Unset,
    'cursor_shape': Unset,
    'font': Unset,
    'allow_bold': Unset,
    'audible_bell': Unset,
    'emulation': Unset,
    'pointer_autohide': Unset,
    'scroll_on_keystroke': Unset,
    'scroll_on_output': Unset,
    'scrollback_lines': 1000,
    'visible_bell': Unset,
    'word_chars': '-A-Za-z0-9,./?%&#:_',
  }

  erase_enum = {
    'auto': vte.ERASE_AUTO,
    'ascii_backspace': vte.ERASE_ASCII_BACKSPACE,
    'ascii_delete': vte.ERASE_ASCII_DELETE,
    'delete_sequence': vte.ERASE_DELETE_SEQUENCE,
    'tty': vte.ERASE_TTY,
  }

  cursor_shape_enum = {
    'block': vte.CURSOR_SHAPE_BLOCK,
    'ibeam': vte.CURSOR_SHAPE_IBEAM,
    'underline': vte.CURSOR_SHAPE_UNDERLINE,
  }

  cursor_blink_enum = {
    'system': vte.CURSOR_BLINK_SYSTEM,
    'on': vte.CURSOR_BLINK_ON,
    'off': vte.CURSOR_BLINK_OFF,
  }

  def __init__(self):
    """Create a new configuration."""
    self.opts = self.default_opts.copy()

  def update(self, options):
    """Update the options from a dict-like or list of tuple-pairs."""
    self.opts.update(options)

  def configure(self, term):
    """Configure the terminal to the stored options."""
    self.set_color_option(term, 'color_foreground')
    self.set_color_option(term, 'color_background')
    self.set_enum_option(term, self.erase_enum, 'backspace_binding')
    self.set_enum_option(term, self.cursor_shape_enum, 'cursor_shape')
    self.set_enum_option(term, self.cursor_blink_enum, 'cursor_blink_mode')
    self.set_font_option(term, 'font')
    self.set_simple_option(term, 'allow_bold')
    self.set_simple_option(term, 'audible_bell')
    self.set_simple_option(term, 'emulation')
    self.set_simple_option(term, 'pointer_autohide')
    self.set_simple_option(term, 'scroll_on_keystroke')
    self.set_simple_option(term, 'scroll_on_output')
    self.set_simple_option(term, 'scrollback_lines')
    self.set_simple_option(term, 'visible_bell')
    self.set_simple_option(term, 'word_chars')

  def set_enum_option(self, term, enum, opt):
    """Set an enum-based option."""
    val = self.opts.get(opt, Unset)
    if val is Unset:
      return
    enumval = enum.get(val, Unset)
    self.set_option(term, opt, enumval)

  def set_font_option(self, term, opt):
    """Set a font option."""
    val = self.opts.get(opt, Unset)
    if val is Unset:
      return
    fontdesc = pango.FontDescription(val)
    self.set_option(term, opt, fontdesc)

  def set_color_option(self, term, opt):
    """Set a color option."""
    val = self.opts.get(opt, Unset)
    if val is Unset:
      return
    color = gtk.gdk.color_parse(val)
    self.set_option(term, opt, color)

  def set_simple_option(self, term, opt):
    """Set a simple option."""
    val = self.opts.get(opt, Unset)
    if val is Unset:
      return
    self.set_option(term, opt, val)

  def set_option(self, term, opt, val):
    """Set an option."""
    if val is Unset:
      return
    setter = getattr(term, 'set_%s' % opt, None)
    if setter is not None:
      setter(val)
    else:
      raise KeyError('bad configuration key "%s".' % opt)


class TerminalView(delegates.SlaveView, lists.ListItem):
  """A8 Terminal Emulator."""

  MARKUP_TEMPLATE = '<b>{0}</b> (<b>{1}</b>)\n<span size="xx-small">{2}</span>'

  @property
  def markup_args(self):
    return (
      os.path.basename(self.cwd),
      self.pid,
      os.path.dirname(self.cwd)
    )

  def create_ui(self):
    self.widget.set_data('delegate', self)
    self.pid = None
    self.cwd = None
    self.box = gtk.HBox()
    self.box.set_spacing(3)
    self.widget.add(self.box)
    self.tools = gtk.VBox()
    self.tools2 = gtk.VBox()
    self.box.pack_start(self.tools, expand=False)
    self.stack = gtk.VBox()
    self.box.pack_start(self.stack)
    self.terminal = vte.Terminal()
    self.config = TerminalConfiguration()
    self.config.update(self.model.config.get('terminal', {}))
    self.config.configure(self.terminal)
    self.add_contexts()
    # Fix the size because by default it is huge.
    self.terminal.set_size_request(50, 50)
    self.stack.pack_start(self.terminal)
    self.scrollbar = gtk.VScrollbar()
    self.scrollbar.set_adjustment(self.terminal.get_adjustment())
    self.box.pack_start(self.scrollbar, expand=False)
    self.box.pack_start(self.tools2, expand=False)
    self.create_tools()
    # TODO(afshar) When 0.26 is in Ubuntu
    #self.create_finder()

  def create_tools(self):
    self.close_button = resources.load_button('cross.png',
      'Close terminal window')
    self.browse_button = resources.load_button('folder.png',
      'Browse working directory')
    self.shell_button = resources.load_button('application_xp_terminal.png',
      'Open a shell in working directory')
    self.bookmark_button = resources.load_button('star.png',
      'Bookmark working directory')
    self.copy_button = resources.load_button('page_white_copy.png',
      'Copy selection')
    self.paste_button = resources.load_button('paste_plain.png',
      'Paste clipboard')
    self.selectall_button = resources.load_button('textfield_add.png',
      'Select all')
    self.selectnone_button = resources.load_button('textfield_delete.png',
      'Select None')
    self.find_button = resources.load_button('find.png',
      'Search for text', gtk.ToggleButton)
    self.tools2.pack_start(self.close_button, expand=False)
    self.tools.pack_start(self.browse_button, expand=False)
    self.tools.pack_start(self.shell_button, expand=False)
    self.tools.pack_start(self.bookmark_button, expand=False)
    self.tools2.pack_start(self.copy_button, expand=False)
    self.tools2.pack_start(self.paste_button, expand=False)
    self.tools2.pack_start(self.selectall_button, expand=False)
    self.tools2.pack_start(self.selectnone_button, expand=False)
    self.copy_button.set_sensitive(False)
    self.create_killer()
    self.create_confirmer()
    # TODO(afshar) When 0.26 is in Ubuntu
    #self.tools.pack_start(self.find_button, expand=False)

  def create_finder(self):
    # TODO(afshar) When 0.26 is in Ubuntu
    self.finder = gtk.HBox()
    self.stack.pack_start(self.finder, expand=False)
    self.finder.set_no_show_all(True)
    self.find_text = gtk.Entry()
    self.findnext_button = resources.load_button('resultset_next.png',
      'Find next')
    self.findprev_button = resources.load_button('resultset_previous.png',
      'Find previous')
    self.finder.pack_start(self.findnext_button, expand=False)
    self.finder.pack_start(self.findprev_button, expand=False)
    self.finder.pack_start(gtk.Label('Regular expression'), expand=False)
    self.finder.pack_start(self.find_text)
    for child in self.finder.get_children():
      child.show()

  def create_confirmer(self):
    self.confirmer = gtk.HBox()
    self.confirmer.set_spacing(3)
    self.confirmer.set_border_width(3)
    self.stack.pack_start(self.confirmer, expand=False)
    self.confirmer.set_no_show_all(True)
    self.confirm_label = gtk.Label()
    self.confirm_label.set_alignment(1, 0.5)
    self.confirmer.pack_start(self.confirm_label)
    self.confirm_yes_button = resources.load_button('tick.png', 'Yes, kill')
    self.confirm_no_button = resources.load_button('cross.png', 'No, save')
    self.confirmer.pack_start(self.confirm_yes_button, expand=False)
    self.confirmer.pack_start(self.confirm_no_button, expand=False)
    for child in self.confirmer.get_children():
      child.show()

  def create_killer(self):
    self.killer_menu = gtk.Menu()
    signums = [
      ('SIGTERM', 15),
      ('SIGKILL', 9),
      ('SIGINT', 2),
      ('SIGABRT', 6),
    ]
    for signame, signum in signums:
      menuitem = gtk.MenuItem()
      menuitem.set_label('{0} ({1})'.format(signame, signum))
      self.killer_menu.append(menuitem)
      menuitem.set_data('signum', signum)
      menuitem.connect('activate', self.on_killer_activate)
    self.killer_menu.show_all()
    self.killer_button = resources.load_button(
      'script_lightning.png', 'Send signal to children')
    self.tools.pack_start(self.killer_button, expand=False)
    self.killer_shell_menu = gtk.Menu()
    for signame, signum in signums:
      menuitem = gtk.MenuItem()
      menuitem.set_label('{0} ({1})'.format(signame, signum))
      self.killer_shell_menu.append(menuitem)
      menuitem.set_data('signum', signum)
      menuitem.connect('activate', self.on_killer_shell_activate)
    self.killer_shell_menu.show_all()
    self.killer_shell_button = resources.load_button(
      'application_lightning.png', 'Send signal to shell')
    self.tools.pack_start(self.killer_shell_button, expand=False)

  def create_tab_widget(self):
    self.tab_box = gtk.HBox()
    self.tab_box.set_spacing(1)
    self.icon = resources.load_icon('application_xp_terminal.png')
    self.tab_box.pack_start(self.icon, expand=False)
    self.label = gtk.Label('Terminal')
    self.label.set_ellipsize(pango.ELLIPSIZE_START)
    small = pango.AttrScale(pango.SCALE_X_SMALL, 0, -1)
    self.label_attributes = pango.AttrList()
    self.label_attributes.insert(small)
    self.label.set_attributes(self.label_attributes)
    self.label.set_width_chars(9)
    self.tab_box.pack_start(self.label)
    self.tab_box.show_all()
    return self.tab_box

  def execute(self, argv=None, env=None, cwd=None):
    if argv is None:
      argv = ['bash']
    if env is None:
      env = self.env_map_to_list(os.environ.copy())
    if cwd is None:
      cwd = os.getcwd()
    log.debug('Start {0} {1}'.format(argv, cwd))
    self.pid = self.terminal.fork_command(argv[0], argv, env, cwd)
    log.debug('Started pid {0}'.format(self.pid))
    self.cwd = cwd
    gobject.timeout_add(500, self.calculate_cwd)

  def env_map_to_list(self, env):
    return ['%s=%s' % (k, v) for (k, v) in env.items()]

  def calculate_cwd(self):
    """Calculate and store the CWD for the running process."""
    if self.pid is None:
      # The return value indicates whether gobject should continue polling.
      return False
    else:
      try:
        self.cwd = psutil.Process(self.pid).getcwd()
        self.label.set_text(self.cwd)
        return True
      except psutil.error.AccessDenied:
        # The process already vanished
        return False

  def add_contexts(self):
    self.contexts = {}
    for context in contexts.ContextManager.context_order:
      self.add_context(context)

  def add_context(self, context):
    matchnum = self.terminal.match_add(context.expr)
    self.contexts[matchnum] = context

  def on_terminal__child_exited(self, terminal):
    self.pid = None
    self.terminal.feed('\x1b[0;1;34mExited, status: \x1b[0;1;31m{0}'.format(
      self.terminal.get_child_exit_status()))

  def on_terminal__window_title_changed(self, terminal):
    pass

  def on_terminal__button_press_event(self, terminal, event):
    col, row = self.get_position_from_pointer(event.x, event.y)
    match = self.terminal.match_check(col, row)
    if match is not None:
      self.on_match(match, event)

  def get_position_from_pointer(self, x, y):
    """Get the row/column position for a pointer position."""
    cw = self.terminal.get_char_width()
    ch = self.terminal.get_char_height()
    return int(x / cw), int(y / ch)

  def on_match(self, match, event):
    match_string, match_index = match
    context_type = self.contexts[match_index]
    context = context_type(self.model, self, match_string)
    log.debug('Matched "{0}" as {1}', match_string, context_type.name)
    if event.button == 3:
      self.on_match_menu(context, event)
    elif event.state & gtk.gdk.CONTROL_MASK:
      self.on_match_default(context, event)

  def on_match_menu(self, context, event):
    menu = context.create_menu()
    if menu is None:
      return
    menu.popup(None, None, None, event.button, event.time)

  def on_match_default(self, context, event):
    pass

  def on_close_button__clicked(self, button):
    if self.pid:
      msg = 'Really? pid{0} is running. We will kill -9'.format(self.pid)
      self.confirm_label.set_text(msg)
      self.confirmer.show()
    else:
      self.close()

  def on_confirm_yes_button__clicked(self, button):
    psutil.Process(self.pid).send_signal(9)
    self.close()

  def on_confirm_no_button__clicked(self, button):
    self.confirmer.hide()

  def close(self):
    self.model.terminals.remove_tab(self)

  def on_browse_button__clicked(self, button):
    self.model.files.browse(self.cwd)

  def on_shell_button__clicked(self, button):
    self.model.terminals.execute(cwd=self.cwd)

  def on_bookmark_button__clicked(self, button):
    self.model.bookmarks.add(self.cwd)

  def on_terminal__selection_changed(self, terminal):
    self.copy_button.set_sensitive(self.terminal.get_has_selection())

  def on_copy_button__clicked(self, button):
    self.terminal.copy_clipboard()

  def on_paste_button__clicked(self, button):
    self.terminal.paste_clipboard()

  def on_selectall_button__clicked(self, button):
    self.terminal.select_all()

  def on_selectnone_button__clicked(self, button):
    self.terminal.select_none()

  def on_killer_button__button_press_event(self, button, event):
    self.killer_menu.popup(None, None, None, event.button, event.time)

  def on_killer_shell_button__button_press_event(self, button, event):
    self.killer_shell_menu.popup(None, None, None, event.button, event.time)

  def on_killer_activate(self, menuitem):
    signum = menuitem.get_data('signum')
    for child in psutil.Process(self.pid).get_children():
      child.send_signal(signum)

  def on_killer_shell_activate(self, menuitem):
    signum = menuitem.get_data('signum')
    psutil.Process(self.pid).send_signal(signum)


class TerminalManager(lists.ListView):
  """Tabs containing terminals."""

  LABEL = 'Terminals'

  def create_ui(self):
    lists.ListView.create_ui(self)
    self.book = gtk.Notebook()
    self.book.set_tab_pos(gtk.POS_BOTTOM)

  def add_tab(self, delegate):
    self.book.append_page(delegate.widget, delegate.create_tab_widget())
    self.book.show_all()
    self.book.set_current_page(self.book.page_num(delegate.widget))
    self.items.append(delegate)

  def remove_tab(self, delegate):
    self.book.remove_page(self.book.page_num(delegate.widget))
    self.items.remove(delegate)

  @property
  def current_page(self):
    return self.book.get_nth_page(self.book.get_current_page())

  @property
  def current_view(self):
    return self.current_page.get_data('delegate')

  def execute(self, argv=None, env=None, cwd=None):
    t = TerminalView(self.model)
    t.execute(argv, env, cwd)
    self.add_tab(t)
    self.model.emit('terminal-executed', argv=argv, env=env, cwd=cwd)

  def on_items__item_activated(self, objectlist, item):
    self.book.set_current_page(self.book.page_num(item.widget))

  def grab_focus(self):
    self.current_view.terminal.grab_focus()

  def next(self):
    current = self.book.get_current_page()
    if current == self.book.get_n_pages() - 1:
      update = 0
    else:
      update = current + 1
    self.book.set_current_page(update)
    self.grab_focus()

  def prev(self):
    current = self.book.get_current_page()
    if current == 0:
      update = self.book.get_n_pages() - 1
    else:
      update = current - 1
    self.book.set_current_page(update)
    self.grab_focus()




# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


import os
import gtk
import yaml

from a8 import lists, contexts, actions


class BookMark(lists.ListItem):

  MARKUP_TEMPLATE = '<b>{0}</b>\n<span size="xx-small">{1}</span>'

  def __init__(self, target):
    self.target = target
    self.basename = os.path.basename(target)
    self.dirname = os.path.dirname(target)

  @property
  def markup_args(self):
    return (self.basename, self.dirname)

  @property
  def type(self):
    print self.target
    if os.path.isdir(self.target):
      return 'dir'
    elif self.target.startswith('http'):
      return 'uri'
    else:
      return 'file'


class BookmarkManager(lists.ListView):

  LABEL = 'Bookmarks'
  ICON = 'star.png'

  remove_action = actions.Action('remove_bookmark', 'Remove Bookmark',
                                 'delete.png')

  def add(self, target):
    self.items.append(BookMark(target))
    self.model.emit('bookmark-item-added', filename=target)
    self.save()

  def create_ui(self):
    lists.ListView.create_ui(self)
    self.filename = self.model.home.path('bookmarks.yaml')
    self.load()

  def load(self):
    self.items.clear()
    if not os.path.exists(self.filename):
      return
    with open(self.filename) as f:
      data = f.read()
    if not data:
      return
    for item in yaml.load(data):
      self.items.append(BookMark(item))

  def shortest_path(self, path):
    match = None
    for item in self.items:
      if item.target in path and (not match or
        len(item.target) > len(match.target)):
        match = item
    return match

  def save(self):
    data = yaml.dump([item.target.encode('utf-8')
                     for item in self.items],
                     default_flow_style=False)
    with open(self.filename, 'w') as f:
      f.write(data)

  def activate(self, bookmark):
    dispatch = {
      'dir': self.activate_dir,
      'file': self.activate_file,
    }
    dispatch[bookmark.type](bookmark)

  def activate_dir(self, bookmark):
    self.model.files.browse(bookmark.target)

  def activate_file(self, bookmark):
    self.model.vim.open_file(bookmark.target)

  def on_items__item_right_clicked(self, items, item, event):
    context = contexts.LocalContext(self.model, None, item.target)
    menu = context.create_menu()
    self.extend_menu(menu)
    menu.show_all()
    menu.popup(None, None, None, event.button, event.time)

  def extend_menu(self, menu):
    menu.append(gtk.SeparatorMenuItem())
    remove_item = self.remove_action.create_menuitem()
    menu.append(remove_item)
    for menuitem in menu.get_children():
      if menuitem.get_data('action_key') == 'bookmark':
        menuitem.set_sensitive(False)
    remove_item.connect('activate', self.on_remove_item_activate)

  def on_remove_item_activate(self, menuitem):
    self.items.remove(self.items.selected_item)
    self.save()

  def on_items__item_activated(self, objectlist, item):
    self.activate(item)




# -*- coding: utf-8 -*- 
# (c) 2005-2012 PIDA Authors
# vim: ft=python sw=2 ts=2 sts=2 tw=80


import logbook
log = logbook.Logger('Extensions')


def import_string(import_name):
  if ':' in import_name:
    module, obj = import_name.split(':', 1)
  elif '.' in import_name:
    items = import_name.split('.')
    module = '.'.join(items[:-1])
    obj = items[-1]
  else:
    return __import__(import_name)
  return getattr(__import__(module, None, None, [obj]), obj)


def load_extension(model, import_name):
  try:
    extension = import_string(import_name)
    extension.setup(model)
    log.debug('Loaded: {0}', import_name)
  except Exception, e:
    log.error('Failed: {0} {1}', import_name, e)


def load_extensions(model):
  extension_names = model.config.get('extensions', [])
  for import_name in extension_names:
    load_extension(model, import_name)
